import re
from dataclasses import dataclass
from typing import Literal, Union


def make_request(obis: str = "", mode: Literal["P", "W", "R"] = "R", data: bytes = b"") -> bytes:
    if type(obis) != str:
        raise TypeError(f"OBIS must be str, not {type(obis).__name__}")

    if mode not in ("P", "W", "R"):
        raise ValueError(f"mode must be in ('P', 'W', 'R'), not '{mode}'")

    if mode in ("P", "W") and not data:
        raise ValueError("data cannot be empty if mode in ('P', 'W')")

    if obis and mode == "P":
        raise ValueError("mode cannot be 'P' if OBIS code was specified")

    if obis:
        pattern = re.compile(r"[A-F0-9]{2}\.[A-F0-9]{2}\.[A-F0-9]{2}\*FF")
        if not bool(pattern.fullmatch(obis)):
            raise ValueError("OBIS code format is wrong")
        obis = obis[:2] + obis[3:5] + obis[6:8] + obis[9:]

    obis = obis.encode()
    data = b"(%s)" % data

    request = b"\x01" + mode.encode() + b"1\x02" + obis + data + b"\x03"
    request += calculate_bcc(request[1:])

    return request


def parse_response(response: bytes) -> Union[str, float, tuple[Union[str, float], ...]]:
    try:
        response = response.split(b"(")[1].split(b")")[0]
    except (IndexError, ValueError):
        raise ValueError(f"invalid response format, response: {response}") from None

    response = response.split(b",")
    if b"." in response[0]:
        if len(response) != 1:
            return tuple(map(float, response))
        return float(response[0])
    if len(response) != 1:
        return tuple(map(lambda x: x.decode("ascii"), response))
    return response[0].decode("ascii")


def calculate_bcc(data: bytes) -> bytes:
    bcc = 0
    for byte in data:
        bcc ^= byte
    return bcc.to_bytes(1, "little")


@dataclass
class Commands:
    transfer_request = b"/?!\r\n"

    # <ACK>0<baudrate_num>1<CR><LF>
    ack = b"\x060%i1\r\n"

    # <SOH>B0<ETX><bcc>
    end_conn = b"\x01B0\x03q"

    total_energy = make_request("0F.08.80*FF")
    voltage_A = make_request("20.07.00*FF")
    voltage_B = make_request("34.07.00*FF")
    voltage_C = make_request("48.07.00*FF")
    active_power_A = make_request("24.07.00*FF")
    active_power_B = make_request("38.07.00*FF")
    active_power_C = make_request("4C.07.00*FF")
    active_power_sum = make_request("10.07.00*FF")
    serial_num = make_request("60.01.00*FF")
    status = make_request("60.05.00*FF")
    season_schedule = make_request("0D.00.00*FF")
    special_days = make_request("0B.00.00*FF")
    tariff_schedule_obis = "0A.%s.64*FF"



from typing import NamedTuple


class Status(NamedTuple):
    data_memory_err: bool
    param_memory_err: bool
    measurement_err: bool
    clock_err: bool
    battery_discharged: bool
    programming_button_pressed: bool
    data_memory_doesnt_work: bool
    param_memory_doesnt_work: bool
    screw_terminal_cover_removed: bool
    load_connected: bool
    load_disconnected: bool


class TotalEnergy(NamedTuple):
    total: float
    T1: float
    T2: float
    T3: float
    T4: float


class Voltage(NamedTuple):
    phaseA: float
    phaseB: float
    phaseC: float


class ActivePower(NamedTuple):
    sum: float
    phaseA: float
    phaseB: float
    phaseC: float


class SeasonSchedule(NamedTuple):
    month: int
    day: int
    weekday_skd_num: int
    sat_skd_num: int
    sun_skd_num: int


class SpecialDaysSchedule(NamedTuple):
    month: int
    day: int
    skd_num: int


class TariffSchedulePart(NamedTuple):
    hour: int
    minute: int
    T_num: int


class TariffSchedule(NamedTuple):
    parts: tuple[TariffSchedulePart, ...]



from time import sleep

import serial
from .types import *
from . import tools


class Meter:
    """
    Custom library for working with electricity meters Neva MT 3xx via serial port.
    """

    BAUDRATES = (300, 600, 1200, 2400, 4800, 9600)
    __init_baudrate = BAUDRATES[0]
    __timeout = 3
    __serial_num = None
    __used_tariff_schedules = set()

    def __init__(self, port):
        self.__port = port
        self.__start_session()

    def __start_session(self):
        """
        Starting serial session, sending initial commands.
        """

        self.session = serial.Serial(port=self.__port,
                                     baudrate=self.__init_baudrate,
                                     bytesize=serial.SEVENBITS,
                                     parity=serial.PARITY_EVEN,
                                     stopbits=serial.STOPBITS_ONE,
                                     timeout=self.__timeout)

        self.send_request(tools.Commands.transfer_request)

        try:
            resp = self.get_response(21)
            working_baudrate_num = int(chr(resp[4]))
            self.__working_baudrate = self.BAUDRATES[working_baudrate_num]
            self.__device_name = resp[5:-2].decode("ascii")
        except IndexError:
            self.close_session()
            raise ConnectionError("identity message format is wrong") from None

        self.send_request(tools.Commands.ack % working_baudrate_num)

        # Flushing data before changing baudrate
        # self.session.flush()
        sleep(.3)
        # Reconfigure port with new baudrate
        self.session.baudrate = self.__working_baudrate

        try:
            self.__password = tools.parse_response(self.get_response(16)).encode()
        except ValueError:
            self.close_session()
            raise ConnectionError("password message format is wrong, try again") from None

        self.send_request(tools.make_request(mode="P", data=self.__password))

        if msg := self.get_response(1) != b"\x06":
            self.close_session()
            raise ConnectionError(f"message is not ACK, message: {msg}")

    @property
    def total_energy(self) -> TotalEnergy:
        """
        Returns сumulative active energy in tariffs T1, T2, T3, T4 and total [kWh].
        """

        self.send_request(tools.Commands.total_energy)
        return TotalEnergy(*tools.parse_response(self.get_response(62)))

    @property
    def voltage_a(self) -> float:
        """Returns instantaneous voltage in phase A (L1) [V]."""

        self.send_request(tools.Commands.voltage_A)
        return tools.parse_response(self.get_response(20))

    @property
    def voltage_b(self) -> float:
        """Returns instantaneous voltage in phase B (L2) [V]."""

        self.send_request(tools.Commands.voltage_B)
        return tools.parse_response(self.get_response(20))

    @property
    def voltage_c(self) -> float:
        """Returns instantaneous voltage in phase C (L3) [V]."""

        self.send_request(tools.Commands.voltage_C)
        return tools.parse_response(self.get_response(20))

    @property
    def voltage(self) -> Voltage:
        """Returns instantaneous voltages of all phases [V]."""

        return Voltage(phaseA=self.voltage_a, phaseB=self.voltage_b, phaseC=self.voltage_c)

    @property
    def active_power_a(self) -> float:
        """Returns active instantaneous power in phase A (L1) [W]."""

        self.send_request(tools.Commands.active_power_A)
        return tools.parse_response(self.get_response(20))

    @property
    def active_power_b(self) -> float:
        """Returns active instantaneous power in phase B (L2) [W]."""

        self.send_request(tools.Commands.active_power_B)
        return tools.parse_response(self.get_response(20))

    @property
    def active_power_c(self) -> float:
        """Returns active instantaneous power in phase C (L3) [W]."""

        self.send_request(tools.Commands.active_power_C)
        return tools.parse_response(self.get_response(20))

    @property
    def active_power_sum(self) -> float:
        """Returns sum of active instantaneous power of all phases [W]."""

        self.send_request(tools.Commands.active_power_sum)
        return tools.parse_response(self.get_response(20))

    @property
    def active_power(self) -> ActivePower:
        """
        Returns active instantaneous power of all phases and their sum [W].
        """

        return ActivePower(sum=self.active_power_sum,
                           phaseA=self.active_power_a,
                           phaseB=self.active_power_b,
                           phaseC=self.active_power_c)

    @property
    def season_schedule(self) -> tuple[SeasonSchedule, ...]:
        self.send_request(tools.Commands.season_schedule)
        zones = tools.parse_response(self.get_response(144))
        zones_parsed = []
        for zone in zones:
            if zone == "0000000000":
                continue
            zone = tuple(int(zone[i:i + 2]) for i in range(0, len(zone), 2))
            self.__used_tariff_schedules.update(zone[2:])
            zones_parsed.append(SeasonSchedule(*zone))
        return tuple(zones_parsed)

    @property
    def special_days_schedule(self) -> tuple[SpecialDaysSchedule, ...]:
        self.send_request(tools.Commands.special_days)
        days = tools.parse_response(self.get_response(236))
        days_parsed = []
        for day in days:
            if day == "000000":
                continue
            day = tuple([int(day[:2]), int(day[2:4]), int(day[4:])])
            self.__used_tariff_schedules.add(day[-1])
            days_parsed.append(SpecialDaysSchedule(*days))
        return tuple(days_parsed)

    @property
    def tariff_schedules(self) -> tuple[TariffSchedule, ...]:
        tariff_schedules = []
        if not self.__used_tariff_schedules:
            self.__used_tariff_schedules.add(1)
        for schedule_num in self.__used_tariff_schedules:
            obis_cmd = tools.Commands.tariff_schedule_obis % f"{schedule_num:02X}"
            self.send_request(tools.make_request(obis_cmd))
            resp = self.get_response(68)
            schedule = tools.parse_response(resp)
            schedule_parsed = []
            for part in schedule:
                if part == "000000":
                    break
                # part = tuple(int(part[i:i + 2]) for i in range(0, len(part), 2))
                part = tuple([int(part[:2]), int(part[2:4]), int(part[4:])])
                schedule_parsed.append(TariffSchedulePart(*part))
            if schedule_parsed:
                tariff_schedules.append(TariffSchedule(tuple(schedule_parsed)))
        return tuple(tariff_schedules)

    @property
    def status(self) -> Status:
        self.send_request(tools.Commands.status)
        response = tools.parse_response(self.get_response(17))
        bits = f'{int(response, 16):0>16b}'
        bits = bits[:8] + bits[9:12]
        return Status(*map(lambda x: x == "1", bits))

    @property
    def serial_number(self) -> str:
        if self.__serial_num:
            return self.__serial_num
        self.send_request(tools.Commands.serial_num)
        self.__serial_num = tools.parse_response(self.get_response(21))
        return self.__serial_num

    @property
    def device_name(self) -> str:
        return self.__device_name

    def send_request(self, request: bytes):
        self.session.write(request)

    def get_response(self, size: int = None, expected: bytes = b"\x03") -> bytes:
        """
        Returns raw bytes response from serial port.
        Without args (by default) it reads up to expected ETX char.
        if size was specified it reads size bytes.
        if expected == None read all before timeout.
        """

        if size:
            return self.session.read(size)
        if expected:
            return self.session.read_until(expected)
        return self.session.readall()

    def disconnect(self):
        self.session.close()

    def connect(self):
        self.session.open()

    def close_session(self):
        if self.session.is_open:
            self.send_request(tools.Commands.end_conn)
            self.session.flush()
            self.session.close()
        self.session.__del__()

    def __str__(self) -> str:
        return self.device_name

    def __repr__(self) -> str:
        return f"Meter({self.__port!r})"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_session()



from .core import Meter
import argparse
from . import tools


def is_property(attr: str) -> bool:
    return type(getattr(Meter, attr)) == property


FUNCTIONS = frozenset(filter(is_property, dir(Meter)))

parser = argparse.ArgumentParser(description="CLI version does not support write mode!\n")

parser.add_argument("--obis", nargs='*', default="", help="Specify OBIS code(-s)")
parser.add_argument("--func", nargs='*', default="", choices=FUNCTIONS,
                    help="Specify prepared function(-s)")
parser.add_argument("-p", "--port", required=True, help="Specify serial port interface")

args = parser.parse_args()

if len(args.obis) == 0 and len(args.func) == 0:
    parser.error("at least one of --obis and --func required")

try:
    with Meter(args.port) as meter:
        print(f"Connected to: {meter}")

        if len(args.obis) != 0:
            print("\nOBIS:")
            for code in args.obis:
                meter.send_request(tools.make_request(code))
                print(f"{code}\t", tools.parse_response(meter.get_response()))

        if len(args.func) != 0:
            print("\nFunctions:")
            for func in args.func:
                print(f"{func}\t", getattr(meter, func))

except ConnectionError as e:
    parser.error(str(e))



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="pyneva",
    version="0.5",
    description="Custom library for electricity meters Neva MT 3xx",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Nikita Nemirovsky",
    author_email="vaze.legend@gmail.com",
    url="https://github.com/vazelegend/pyneva",
    license="MIT",
    packages=find_packages(),
    install_requires=["pyserial>=3.5"],
    python_requires=">=3.9",
    platforms="any",
)
